/* krb5sia.c
 *
 * Luke-karol Petelczyc - luke@cit.uws.edu.au
 *
 * This is a plugin module for Compaq's SIA system. It will allow you to 
 * authenication users against a kerberos server.
 *
 *
 * See for list of SIA functions
 * http://www.tru64unix.compaq.com/docs/base_doc/
 * => DOCUMENTATION/HTML/AA-Q0R2D-TET1_html/sec.c224.html
 *
 *
 * To Compile
 * cc -I/usr/local/include -c krb5sia.c
 * cc -shared -o libkrb5sia.so krb5sia.o -lc \
 *    -L/share/src/luke/krb5-1.2.2/src/lib/ \
 *    -lkrb5 -lk5crypto -lcom_err
 *
 * To Install
 * siacfg -a KRB5 /usr/shlib/libkrb5sia.so
 */

/* LICENSE
Copyright (c) 2002, Luke-karol Petelczyc
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

	Redistributions of source code must retain the above copyright notice,
	this list of conditions and the following disclaimer. 

	Redistributions in binary form must reproduce the above copyright 
	notice, this list of conditions and the following disclaimer in 
	the documentation and/or other materials provided with the 
	distribution. 

	Neither the name of the University of Western Sydney nor the names
	of its contributors may be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "krb5sia.h"

/* Only performed once per reboot( read security system starting) */
int 
siad_init( void )
{
  SIALOG("DEBUG","siad_init <- luke");
  return SIADSUCCESS;
}

/* Verify the calling programs privileges */
int
siad_chk_invoker( void )
{
  SIALOG("DEBUG","siad_chkinvoker <- luke");
  return SIADFAIL;
}



/*****************************************************************
 **  Session Functions
 *****************************************************************/

/* Initialise the session */
int
siad_ses_init( SIAENTITY *entity, int pkgind )
{
  return krb5_siad_ses_init( entity, pkgind );
}




/* Authenticate the session */
int
siad_ses_authent( sia_collect_func_t *collect,
		  SIAENTITY *entity,
		  int siastat,
		  int pkgind )
{
  SIALOG("DEBUG","siad_ses_authent()");
  return krb5_siad_ses_authent( collect, entity, siastat, pkgind );
}




/* Check resources and licences */
int
siad_ses_estab( sia_collect_func_t *collect,
		SIAENTITY *entity,
		int pkgind )
{
  SIALOG("DEBUG","siad_ses_estab()");
  return SIADFAIL;
}



/* Push cache details out into users environment */
int
siad_ses_launch( sia_collect_func_t *collect,
		  SIAENTITY *entity,
		  int pkgind )
{
  return krb5_siad_ses_launch( collect, entity, pkgind );
}



/* Process the 'su' command */
int
siad_ses_suauthent(sia_collect_func_t *collect,
		   SIAENTITY *entity,
		   int siastat,
		   int pkgind )
{
  SIALOG("DEBUG","siad_ses_suauthent()");
  return krb5_siad_ses_suauthent( collect, entity, siastat, pkgind );
}



/* Re-validate password */
int 
siad_ses_reauthent( sia_collect_func_t *collect,
		    SIAENTITY *entity,
		    int siastat,
		    int pkgind )
{
  return krb5_siad_ses_reauthent( collect, entity, siastat, pkgind );
}

/* Release sessions resources */
int
siad_ses_release( SIAENTITY *entity, 
		  int pkgind )
{
  return krb5_siad_ses_release( entity, pkgind );
}
 




/*****************************************************************
 **  Change Functions
 **
 **  We support none of these as they should all be done by LDAP
 *****************************************************************/


/* Process the chfn command */
int
siad_chg_finger( sia_collect_func_t *collect,
		 const char *username,
		 int argc,
		 char *argv[])
{
  SIALOG("DEBUG","siad_chfn <- luke");
  return SIADFAIL;
}

/* process a change password */
int 
siad_chg_password( sia_collect_func_t *collect,
		   const char *username,
		   int argc,
		   char *argv[])
{
  SIALOG("DEBUG","siad_chg_passwd()");
  return krb5_siad_chg_password( collect, username, argc, argv );
}

/* process a chsh */
int
siad_chg_shell( sia_collect_func_t *collect,
		const char *username,
		int argc,
		char *argv[])
{
  SIALOG("DEBUG","siad_shell <- luke");
  return SIADFAIL;
}





int 
siad_chk_user( const char *logname, 
	       int checkflag )
{
  return krb5_siad_chk_user( logname, checkflag );
}

int
siad_get_groups( struct sia_context *context, 
		 const char *username,
		 gid_t *buf, 
		 int *numgroups,
		 int maxgroups )
{
  SIALOG("DEBUG","siad_get groups <- luke");
  return SIADFAIL;
}




/*****************************************************************
 ** krb5_siad_* functions
 ** These are called by the siad_ functions
 *****************************************************************/

/* initialise the krb5 context and store the address in the 
 * entity mechanism table */
int
krb5_siad_ses_init( SIAENTITY *entity, int pkgind )
{
  krb5_error_code retVal;
  // look at __madvisor to turn off swapping
  struct siakrb5ses *ses = malloc( sizeof( struct siakrb5ses ) );

  /* check that malloc worked, and then zero mem */
  if( NULL == ses )
    return SIADFAIL;

  memset( ses, 0, sizeof( struct siakrb5ses ) );
  ses->authedOK=0;

  retVal = krb5_init_context((krb5_context *)&ses->context );
  if( retVal )
    {
      free( ses );
      return SIADFAIL;
    }
  
  /* Everything worked, store the location of the session details. */
  entity->mech[ pkgind ] = (int *)ses;
  return SIADSUCCESS;
}



/* authentcate user */
int
krb5_siad_ses_authent( sia_collect_func_t *collect,
		       SIAENTITY *entity,
		       int siastat,
		       int pkgind )
{
  return krb5_auth(collect, entity, siastat, pkgind );
}

/* Re-authenticate user. */
/* It's important to note that this function does not create a cache,
   it simply returns success or fail */
int 
krb5_siad_ses_reauthent( sia_collect_func_t *collect,
			 SIAENTITY *entity,
			 int siastat,
			 int pkgind )
{
  SIALOG("DEBUG","siad_ses_reauthent()");
  if( NULL == entity )
    return SIADFAIL;
  
  return krb5_auth( collect, entity, siastat, pkgind );
}


/* Launch session */
int
krb5_siad_ses_launch( sia_collect_func_t *collect,
		      SIAENTITY *entity,
		      int pkgind )
{
  return krb5_cache_config( entity, pkgind );
}



/* Release all memory we've used back to OS */
int
krb5_siad_ses_release( SIAENTITY *entity, int pkgind )
{
  struct siakrb5ses *s = (struct siakrb5ses *) entity->mech[pkgind];
  
  krb5_free_principal( (krb5_context)s->context, s->princ );
  krb5_free_context( (krb5_context)s->context );
  free( s );
  entity->mech[pkgind] = (int *)0;

  SIALOG("DEBUG","siad_ses_release(%s)", entity->name);
  return SIADSUCCESS;
}







int
krb5_auth(sia_collect_func_t *collect,
	  SIAENTITY *entity,
	  int siastat,
	  int pkgind )
{
  prompt_t prompts[2];
  prompt_t *prompt;
  char *uname = (char *)0;

  /* If user has already been authenticated by higher layer leave */
  /* ie, this SIA plugin supports SIA LAYER VOUCHING */
  if( SIADSUCCESS == siastat && geteuid() == 0 )
    {
      SIALOG( "DEBUG", "Vouched!");
      return SIADSUCCESS;
    }

  /* check that the entity is there, else fail */
  if( NULL == entity )
    {
      SIALOG("DEBUG", "ENTITY==NULL");
      return (SIADFAIL | SIADSTOP);
    }

  uname = entity->name;
  if( entity->acctname )
    uname = entity->acctname;

  /* is there a supplied collection function, and we can collect input */
  /* NOTE! Tru64 Documentation calls this entity->can_collect_input
           How poor that they can't keep their documentation accurate! */
  if( NULL != collect && entity->colinput )
    {
      int n = 0;
      prompt = prompts;
      /* Create prompts to get username and password and allocate memory */
      if( NULL == uname )
	{
	  if( SIADSUCCESS != krb5_get_uname( entity, prompt )  )
	    return SIADFAIL;
	  prompt++;
	}

      if( NULL == entity->password  )
	{
	  if( SIADSUCCESS != krb5_get_passwd( entity, prompt ) )
	    return SIADFAIL;
	  prompt++;
	}
      
      /* we made n prompts */
      n = prompt - prompts;

      /* call the collection functions with number of prompts required */
      if((*collect)( 0,((n>1)?SIAFORM:SIAONELINER),(unsigned char*)"",n,prompts) !=
	 SIACOLSUCCESS )
	return SIADFAIL | SIADSTOP;

    }

  /* Collected details, check for errors */
  
  /* is name null?? */
  if( NULL == uname )
    {
      SIALOG("DEBUB","if( NULL==uname )");
      uname = entity->name;
    }
  if( NULL == uname || '\0' == uname[0] )
    {
      SIALOG("BEDUG","  if( NULL == uname || '\0' == uname[0] )");
      return SIADFAIL;
    }
  
  /* is there a password, REMEMBER , password can be null */
  if( NULL == entity->password || strlen( entity->password ) > SIAMXPASSWORD )
    return SIADFAIL;


  SIALOG("SUCCESS","krb5_siad_ses_authent");
  
  return krb5_authenticate( entity, pkgind );
  
}


int
krb5_get_uname( SIAENTITY *e, prompt_t *p )
{
  SIALOG("DEBUG", "krb5_setup_uname");

  e->name = malloc( SIANAMEMIN + 1 );
  if( NULL == e->name )
    return SIADFAIL;
  
  /* Create username prompt and assign result to name field */
  p->prompt = (unsigned char *)"Login: ";
  p->result = (unsigned char *) e->name;
  p->min_result_length = 1;
  p->max_result_length = SIANAMEMIN;
  p->control_flags = 0;
  return SIADSUCCESS;
}

int 
krb5_get_passwd( SIAENTITY *e, prompt_t *p )
{
  SIALOG("DEBUG", "krb5_setup_passwd");

  e->password = malloc( SIAMXPASSWORD + 1 );
  if( NULL == e->password )
    return SIADFAIL;
  
  /* Create password prompt and assign result to password field */
  p->prompt = (unsigned char *)"Password: ";
  p->result = (unsigned char *) e->password;
  p->min_result_length = 0;
  p->max_result_length = SIAMXPASSWORD;
  p->control_flags = SIARESINVIS;
  return SIADSUCCESS;
}



/* Authenticate username against password, and create ccache */
int
krb5_authenticate( SIAENTITY *entity, 
		   int pkgind )
{

  krb5_creds creds;                 /* credentials */
  krb5_get_init_creds_opt options;  /* credential options */
  struct passwd pwd;
  struct passwd *p_pwd;
  char passbuf[SIABUFSIZ];

  /* retrieve our session details */
  struct siakrb5ses *s = (struct siakrb5ses *) entity->mech[pkgind];
  
  /* check we have a username by now, else we need to fail! */
  if( NULL == entity->name )
    {
      SIALOG("DEBUG","e->name==NULL <- luke");
      return SIADFAIL;
    }

  /* create a passwd structure for this entity */
  if( getpwnam_r( entity->name, &pwd, passbuf, sizeof(passbuf), &p_pwd ) != 0 )
    {
      SIALOG("ERROR","getpwnam_r()");
      return SIADFAIL;
    }

  /* copy password structure into entity */
  if( sia_make_entity_pwd(&pwd, entity) == SIADFAIL )
    {
      SIALOG("ERROR","sia_make_entity_pwd()");
      return SIADFAIL;
    }
  
  /* Create a filename for the credentials cache */
  memset( s->cachefile, 0, PATH_MAX );
  snprintf( s->cachefile, PATH_MAX, "FILE:/tmp/krb5cc%d%d",
	    pwd.pw_uid, getpid() );
  SIALOG("HELP", s->cachefile );

  /* convert username to a krb5 principal */
  if( krb5_parse_name( (krb5_context)s->context, entity->name, &s->princ ) )
    {
      SIALOG("DEBUG","krb5_parse_name <- luke");
      return SIADFAIL;
    }
  
  /* create the credentials cache */
  if( krb5_cc_resolve((krb5_context) s->context, s->cachefile, &s->cc ) )
    {
      SIALOG("DEBUG","krb5_cc_default()");
      return SIADFAIL;
    }

  
  /* alloc memory for our options and credentials */
  krb5_get_init_creds_opt_init( &options );
  memset( &creds, 0, sizeof(creds) );

  /* dont adjust options, users _WILL_ be happy with our defaults*/
  if( krb5_get_init_creds_password((krb5_context) s->context,
				    &creds,
				    s->princ,
				    entity->password,
				    (void *)0, /* password prompt */
				    0,         /* data? */
				    0, /* starttime */
				    (void *)0, /* service */
				    &options ) ) /* options :) */
    {
      SIALOG("DEBUG","AUTH FOR %s AGAINST KRB5 FAILED!", entity->name );
      SIALOG("DEBUG","krb_get_init_creds_password");
      return SIADFAIL;
    }

  /* trash the password from memory so if we crash or something, no one */
  /* else can read it */
  memset( entity->password, 0, strlen(entity->password) );
  s->authedOK = 1;

  SIALOG("INFO","AUTHENTICATED %s TO KERBEROS!!!", entity->name);

  /* Initialise the cache for this principal */
  if( krb5_cc_initialize((krb5_context)s->context, s->cc, s->princ ) )
    {
      SIALOG("DEBUG","krb5_cc_initialize()");
      return SIADFAIL;
    }

  /* Store credentials in the cache we have made */
  if( krb5_cc_store_cred((krb5_context) s->context, s->cc, &creds ) )
    {
      SIALOG( "DEBUG","krb5_cc_store_cred()");
      return SIADFAIL;
    }
  

  SIALOG( "INFO", "Created cache, try klist" );
  return SIADSUCCESS;
}




/* Make sure that we prepare the ownership on the cache file and also
   push the environment variables into the users environment */
int
krb5_cache_config( SIAENTITY *entity, int pkgind )
{
  struct siakrb5ses *s = (struct siakrb5ses *) entity->mech[pkgind];


  if( s->authedOK )
    {
      char *env = (void *)0;
      int len = (sizeof(char)*PATH_MAX)+(strlen(KRB5CCNAME));


      /* Place the environment variable onto the envrionment list */
      env = malloc( len+1 );
      if( NULL == env )
	{
	  SIALOG("ERR","malloc");
	  return SIADFAIL;
	}

      memset( env, 0, len );
      if( snprintf( env, len, "%s%s", KRB5CCNAME, s->cachefile ) <= 0 )
	{
	  SIALOG("ERR","snprintf");
	  return SIADFAIL;
	}
      
      if( putenv( env ) != 0 )
	{
	  SIALOG("ERR", "putenv" );
	  return SIADFAIL;
	}
      
      
      /* change ownership of files to user about to login. */
      if( chown( &s->cachefile[ sizeof("FILE:") -1 ], entity->pwd->pw_uid,
		 entity->pwd->pw_gid ) != 0 )
	{
	  SIALOG("ERR","%s", &s->cachefile[ sizeof("FILE:")-1 ] );
	  return SIADFAIL;
	}
      
      /* Although it feels right to free env, we musn't according to the 
	 manual entry for putenv
	 free( env );
      */
    }
  
  return SIADSUCCESS;
}


int
krb5_siad_chg_password( sia_collect_func_t *collect, 
		   const char *username,
		   int argc, 
		   char **argv )
{
  krb5_context ctx;
  krb5_principal princ;
  krb5_get_init_creds_opt opts;
  krb5_creds creds;
  krb5_data res_code_str,
    res_str;
  int res_code=0;
  prompt_t prompts[6];

  unsigned char *oldpw;
  unsigned char *newpw1;
  unsigned char *newpw2;
  unsigned char temp[1];


  /* Check that a username is available, if not exit */
  if( NULL == username )
    {
      SIALOG("DEBUG","NULL==username" );
      return SIADFAIL;
    }

  /* Allocate and zero memory for passwords */
  oldpw = malloc( sizeof( char ) * SIAMXPASSWORD + 1 );
  newpw1 = malloc( sizeof( char ) * SIAMXPASSWORD + 1 );
  newpw2 = malloc( sizeof( char ) * SIAMXPASSWORD + 1 );

  if( NULL == oldpw || NULL == newpw1 || NULL == newpw2 )
    {
      SIALOG("ERROR","malloc()");
      if( NULL != oldpw ) free( oldpw );
      if( NULL != newpw1 ) free( newpw1 );
      if( NULL != newpw2 ) free( newpw2 );
      return SIADFAIL;
    }

  memset( oldpw, 0, SIAMXPASSWORD+1 );
  memset( newpw1, 0, SIAMXPASSWORD+1 );
  memset( newpw2, 0, SIAMXPASSWORD+1 );


  /* initialize kerberos library */
  if( krb5_init_context( &ctx ) )
    {
      SIALOG( "DEBUG", "krb5_chg_password(); krb5_init_context()" );
      free( oldpw );
      free( newpw1);
      free( newpw2);
      return SIADFAIL;
    }

  if( krb5_parse_name( ctx, username, &princ ) )
    {
      SIALOG( "DEBUG", "krb5_chg_password(); krb5_parse_name()");
      free( oldpw );
      free( newpw1);
      free( newpw2);
      krb5_free_context( ctx );
      return SIADFAIL;
    }

  /* set up all the prompts */
  prompts[0].prompt = (unsigned char *) "Old Password: ";
  prompts[0].result = oldpw;
  prompts[0].min_result_length = 0;
  prompts[0].max_result_length = SIAMXPASSWORD;
  prompts[0].control_flags = SIARESINVIS;

  prompts[1].prompt = (unsigned char *) "New Password: ";
  prompts[1].result = newpw1;
  prompts[1].min_result_length = 0;
  prompts[1].max_result_length = SIAMXPASSWORD;
  prompts[1].control_flags = SIARESINVIS;

  prompts[2].prompt = (unsigned char *) "New Again: ";
  prompts[2].result = newpw2;
  prompts[2].min_result_length = 0;
  prompts[2].max_result_length = SIAMXPASSWORD;
  prompts[2].control_flags = SIARESINVIS;

  prompts[3].prompt = (unsigned char *) "Incorrect Password :(";
  prompts[3].result = &temp[0];
  prompts[3].min_result_length = 0;
  prompts[3].max_result_length = 0;
  prompts[3].control_flags = SIARESINVIS;
  
  prompts[4].prompt = (unsigned char *) "Cannot authenticate service :(";
  prompts[4].result = &temp[0];
  prompts[4].min_result_length = 0;
  prompts[4].max_result_length = 0;
  prompts[4].control_flags = SIARESINVIS;

  prompts[5].prompt = (unsigned char *) "Passwords don't match :(";
  prompts[5].result = &temp[0];
  prompts[5].min_result_length = 0;
  prompts[5].max_result_length = 0;
  prompts[5].control_flags = SIARESINVIS;

  /* Collect passwords from user */
  if( (*collect)( 120, SIAFORM, NULL, 3, prompts ) != SIACOLSUCCESS )
    {
      SIALOG( "ERROR", "FAiled to collect anything,,.,");
      /* release context and principal here */
      free( oldpw );
      free( newpw1);
      free( newpw2);
      krb5_free_principal( ctx, princ );
      krb5_free_context( ctx );
      return SIADFAIL;
    }

  
  /* set defaults */
  krb5_get_init_creds_opt_init( &opts );
  krb5_get_init_creds_opt_set_tkt_life( &opts, (5*60) );
  krb5_get_init_creds_opt_set_renew_life( &opts, 0 );
  krb5_get_init_creds_opt_set_forwardable( &opts, 0 );
  krb5_get_init_creds_opt_set_proxiable( &opts, 0 );


  /* Get credentials to change the password */
  if( krb5_get_init_creds_password( ctx,        /* context */
				    &creds,     /* credential cache */
				    princ,      /* principal */
				    (char *)oldpw,/* the current password */
				    NULL,       /* function to get password */
				    NULL,       /* data for the password func */
				    0,          /* starttime of tickets */
				    "kadmin/changepw", /* name of service */
				    &opts ) )   /* options */ 
    {
      /* display error message */
      (* collect)(0,SIAWARNING,(unsigned char *)"Kerberos",1, &prompts[3] );
      SIALOG( "DEBUG", "krb5_get_init_creds_password() " );
      free( oldpw );
      free( newpw1);
      free( newpw2);
      krb5_free_principal( ctx, princ );
      krb5_free_context( ctx );
      return SIADFAIL;
    }


  /* check both the collected passwords are the same */
  if( strncmp( (const char *)newpw1, (const char *)newpw2, MAX(strlen((const char *)newpw1), strlen((const char * )newpw2)) ) != 0 )
    {
      (* collect)(0,SIAWARNING,(unsigned char *)"Kerberos",1,&prompts[5] );
      free( oldpw );
      free( newpw1);
      free( newpw2);
      krb5_free_principal( ctx, princ );
      krb5_free_context( ctx );
      SIALOG("DEBUG", "Passwords dont match" );
      return SIADFAIL;
    }


  /* change the users password */
  if( krb5_change_password( ctx,           /* context */
			    &creds,        /* credentials */
			    (char *)newpw1,/* new password */
			    &res_code,     /* result code */
			    &res_code_str, /* result code string */
			    &res_str ) )   /* result string */
    {
      (* collect)(0,SIAWARNING,(unsigned char *)"Kerberos",1, &prompts[4] );
      SIALOG( "DEBUG", "krb5_change_password()" );
      free( oldpw );
      free( newpw1);
      free( newpw2);
      krb5_free_principal( ctx, princ );
      krb5_free_context( ctx );
      return SIADFAIL;
    }

  /* return our stuff to the OS */
  free( oldpw );
  free( newpw1);
  free( newpw2);
  krb5_free_principal( ctx, princ );
  krb5_free_context( ctx );


  /* Password changed ok */
  SIALOG("INFO","krb5_change_password() OK!");
  return SIADSUCCESS;
}


/* we only want to return success if we are being requested to do a 
   change password */
int 
krb5_siad_chk_user( const char *logname, 
		    int checkflag )
{
  SIALOG("DEBUG", "krb5_chk_user(%s)!!!!",logname);
  return ( CHGPASSWD == checkflag ) ? SIADSUCCESS : SIADFAIL;
}



/* See if we can switch user to this new one. */
int
krb5_siad_ses_suauthent( sia_collect_func_t *collect,
			 SIAENTITY *entity,
			 int siastat,
			 int pkgind )
{
  struct passwd pwd;
  struct passwd *p_pwd;
  char passbuf[SIABUFSIZ];

  /* Accept vouching for su */
  if( siastat && geteuid() == 0 )
    return SIADSUCCESS;

  /* is our effective UID 0? it should be */
  if( geteuid() != 0 )
    return SIADFAIL;

  /* is root running this command? */
  if( getuid() == 0 )
    {
      /* then dont worry about getting a password */
      if( getpwnam_r( entity->name, &pwd, passbuf, sizeof(passbuf), &p_pwd ) != 0 )
        {
          SIALOG("ERROR","getpwnam_r()");
          return SIADFAIL;
        }

      /* copy password structure into entity */
      if( sia_make_entity_pwd(&pwd, entity) == SIADFAIL )
        {
          SIALOG("ERROR","sia_make_entity_pwd()");
          return SIADFAIL;
        }

      return SIADSUCCESS;
    }

  /* Is any of the entity stuff null????? */
  if( NULL == entity )
    return SIADFAIL;

  if( NULL == entity->name )
    return SIADFAIL;

  /* if not null ptr, but first char is \0, change name to root */
  if( '\0' == entity->name[0] )
    { 
      /* free the name, and then strdup will alloc new space */
      free( entity->name );
      entity->name = strdup("root");

      /* did strdup work? */
      if( NULL == entity-> name )
	return SIADFAIL;
    }

  /* Did they get the password right? */
  if( SIADFAIL == krb5_auth( collect, entity, pkgind, siastat ) )
    return SIADFAIL;

  /* Password was ok, lets create the credentials cache */
  return krb5_cache_config( entity, pkgind );
}
